// CallocInflable.cpp
//
// Copied (with possible minor changes) from:
// http://www.codeguru.com/Cpp/Cpp/cpp_mfc/buffers/article.php/c4113
// Inflable Arrays in Win32
// Petko Popov (view profile)
// September 21, 2001

#include "stdafx.h"
#include "CallocInflatable.h"

__int32 UTL_CommitMemory(LPEXCEPTION_POINTERS ep,
                         void * pBaseAddress,
                         void * pLastAddress,
                         size_t dwElementSize)
{
   PVOID    pVirtualAddress =
       (PVOID)(ep->ExceptionRecord->ExceptionInformation[1]);

   if (  pVirtualAddress <  pBaseAddress ||
         pVirtualAddress >= pLastAddress) {
      // well, access violation is not within the array bounds -
      //  don't mess up with it
      return EXCEPTION_CONTINUE_SEARCH;
   }

   // we must figure out where the element starts and where it ends
   // otherwise we may get in heavy trouble if first accessing
   // the non-first byte of the last element
   // this is obvious when the element size is large, but it is also
   // dangerous for small sizes

   // figure out what to allocate
   void *   pBaseAddressOfAccessedElement =
            (BYTE *)pBaseAddress +
            (((BYTE *)pVirtualAddress -
                (BYTE *)pBaseAddress) / dwElementSize) * dwElementSize;

   // Memory allocated by VirtualAlloc is automatically initialized
   // to zero, unless MEM_RESET is specified
   if ( ::VirtualAlloc(pBaseAddressOfAccessedElement,
                       dwElementSize,
                       MEM_COMMIT,
                       PAGE_READWRITE) == NULL ) {
      return EXCEPTION_CONTINUE_SEARCH;
   } else {
      return EXCEPTION_CONTINUE_EXECUTION;
   }
}

void UTL_DecommitMemory(void * pBaseAddress,
                        size_t dwTotalSize)
{
   if ( pBaseAddress != NULL && dwTotalSize > 0 ) {
      VirtualFree(pBaseAddress, dwTotalSize, MEM_DECOMMIT);
   }
}

void UTL_ReleaseMemory(void * pBaseAddress,
                       size_t dwTotalSize)
{
   if ( pBaseAddress != NULL ) {
      UTL_DecommitMemory(pBaseAddress, dwTotalSize);
      VirtualFree(pBaseAddress, 0, MEM_RELEASE);
   }
}

