// LispGC.cpp : Defines the entry point for the console application.
//

#include "stdafx.h"
#include "CallocInflatable.h"
#include "MasterPointer.h"
#include "ConsPair.h"
#include "MemoryRegion.h"
#include "RootPointer.h"
#include "MutableList.h"
#include "LispGC.h"

static MemoryRegion Space1;
static MemoryRegion Space2;

MutableList RootList;
const RootPointer Null;

static MemoryRegion * FromSpace;
static MemoryRegion * ToSpace;
static size_t indexToSpace = 0;

size_t PageSize = 0;

ConsPair * MasterToCons(MasterPointer * ptr)
{
	if (ptr == NULL)
	{
		return NULL;
	}
	return ptr->m_ptr;
}

static bool RootScanRunning()
{
	return RootList.Current() != NULL;
}

static bool ObjectScanRunning()
{
	return indexToSpace < ToSpace->GetUsed();
}

static void Swap()
{
	RootList.Start();
	indexToSpace = 0;
	MemoryRegion * temp = ToSpace;
	ToSpace = FromSpace;
	FromSpace = temp;
	ToSpace->Clear();
}

static ConsPair * CopyObject(ConsPair * consPair)
{
	if (ToSpace->Belongs(consPair))
	{
		return consPair;
	}
	ConsPair * result = ToSpace->Cons();
	consPair->m_redirection.m_ptr = result;
	* result = * consPair;
	return result;
}

static void ScanObject(ConsPair * consPair)
{
	// Copy the car and cdr's:
	ConsPair * car = MasterToCons(consPair->m_car);
	ConsPair * cdr = MasterToCons(consPair->m_cdr);
	if (car != NULL)
	{
		// Copy the object pointed to,
		// then fix the pointer itself
		// to point to the new location.
		car = CopyObject(car);
		consPair->m_car = &(car->m_redirection);
	}
	if (cdr != NULL)
	{
		// Copy the object pointed to,
		// then fix the pointer itself
		// to point to the new location.
		cdr = CopyObject(cdr);
		consPair->m_cdr = &(cdr->m_redirection);
	}

}

void ScanPointer(RootPointer * rootPointer)
{
	// Copy the object pointed to,
	// then fix the pointer itself
	// to point to the new location.
	ConsPair * consPair = MasterToCons(rootPointer->m_ptr);
	if (consPair != NULL)
	{
		consPair = CopyObject(consPair);
		rootPointer->m_ptr = &(consPair->m_redirection);
		// Un-Comment this for
		// more aggressive scanning but
		// somewhat slower execution:
		//
		// ScanObject(consPair);
	}
}

static void ScanPointer()
{
	if (RootScanRunning())
	{
		ScanPointer(RootList.Current());
		RootList.Next();
	}
}

static void ScanObject()
{
	if (ObjectScanRunning())
	{
		ScanObject(ToSpace->Get(indexToSpace));
		++indexToSpace;
	}
}

bool GCAdvance()
{
	// Pointer Scan:
	ScanPointer();
	// Object Scan:
	ScanObject();
	// Re-Start:
	if ((!RootScanRunning()) && (!ObjectScanRunning()))
	{
		Swap();
		return true;
	}
	return false;
}

void GCCollect()
{
	// Finish the current GC cycle:
	while (!GCAdvance()) {}
	// Then sweep anything left:
	while (!GCAdvance()) {}
	cout <<
		"Collected, used of FromSpace: " << 
		FromSpace->GetUsed() << endl;
}

void GCDiagnostics()
{
	cout << "Used in FromSpace: " << FromSpace->GetUsed() << endl;
	cout << "Used in ToSpace: " << ToSpace->GetUsed() << endl;
}

RootPointer Cons(const RootPointer & car, const RootPointer & cdr)
{
	GCAdvance();
	RootPointer rootPointer;
	ConsPair * result = ToSpace->Cons();
	memset(result, 0, sizeof(ConsPair));
	result->m_car = car.m_ptr;
	result->m_cdr = cdr.m_ptr;
	result->m_redirection.m_ptr = result;
	rootPointer.m_ptr = &(result->m_redirection);
	return rootPointer;
}

static void FindPageSize()
{
	SYSTEM_INFO systemInfo;
	ZeroMemory(&systemInfo, sizeof (SYSTEM_INFO));
	GetSystemInfo(&systemInfo);
	PageSize = systemInfo.dwPageSize;
}

size_t GetPageSize()
{
	return PageSize;
}

size_t RoundUpToPageSize(size_t size)
{
	size_t multipleOf = PageSize * 2;
	if ((size % multipleOf) == 0)
	{
		return size;
	}
	return ((size / multipleOf) * multipleOf) + multipleOf;
}

static int SetupGC(int argc, _TCHAR* argv[])
{
	int result = 1;
	BYTE * memory1;
	BYTE * memory2;
	const size_t theSize = RoundUpToPageSize(
		size_t(1024) * 1024 * 1000);
	cout << "Will allocate " << theSize << " bytes." << endl;
	BYTE * memory;

	VIRTUAL_CALLOC(memory, theSize);
	// cause an exception if the memory pointer is NULL:
	memory[0] = 0;

	size_t singleSize = theSize / 2;
	memory1 = memory;
	memory2 = memory + singleSize;

	Space1.SetData(memory1, singleSize);
	Space2.SetData(memory2, singleSize);

	FromSpace = &Space1;
	ToSpace = &Space2;

	GCAdvance();

	result = gcmain(argc, argv);

	VIRTUAL_FREE
	return result;
}

int _tmain(int argc, _TCHAR* argv[])
{
	FindPageSize();
	int result = SetupGC(argc, argv);
	return result;
}

