#include "stdafx.h"

#include "RootPointer.h"
#include "LispGC.h"

#include <time.h>

void pause()
{
	GCDiagnostics();
	printf("Enter 0 to continue: ");
	int x;
	scanf_s("%d", &x);
}

void StressTest(RootPointer & a)
{
	a = Cons(a, a);
	for (int i = 0; i < 1000000; ++i)
	{
		a =
			Cons(
				Null,
				a.SetCar(
					Cons(
						a.Car(),
						Cons(a.Cdr(), a))));
	}
}

void IterationTest(RootPointer & a)
{
	for (int i = 0; i < 1000000; ++i)
	{
		a = Cons(a, Null);
		RootPointer b = Cons(a, a);
		a.SetCdr(b);
		a = Cons(a, b);
		b = Cons(b, a);
		a = Cons(a, b);
		b = Cons(b, a);
		a = Cons(a, b);
		b = Cons(b, a);
		a = Cons(a, b);
		b = Cons(b, a);
		a = Cons(a, b);
		b = Cons(b, a);
		a = Cons(a, b);
		b = Cons(b, a);
		a = Cons(a, b);
		b = Cons(b, a);

		// Create a cross-link:
		a = Cons(Null, a);
		b = Cons(Null, b);
		a.SetCar(b);
		b.SetCar(a);

		// Lose the pointer:
		a = Null;
	}
}

#define ExecuteTimer(Function, Pointer) \
	Timer((Function), (Pointer), (#Function))

void Timer(
	void (*Function)(RootPointer &),
	RootPointer a,
	char * description
)
{
	time_t first, last;
	time(&first);
	Function(a);
	time(&last);
	time_t diff = last - first;
	cout << description << " finished in " << diff << " seconds." << endl;
}

int gcmain(int argc, _TCHAR* argv[])
{
	UNREFERENCED_PARAMETER(argc);
	UNREFERENCED_PARAMETER(argv);

	RootPointer a = Cons(Null, Null);
	a.Int() = 10;
	a.Double() = 3.1415;
	{
		RootPointer b = Cons(a, Null);
		RootPointer c = Cons(b, a);
		b.SetCdr(c);
		cout << "c.Cdr().Int() = " << c.Cdr().Int() << endl;
	}
	{
		RootPointer b = Cons(a, Null);
		RootPointer c = Cons(b, a);
		b.SetCdr(c);
		a.SetCdr(b);
		cout << "c.Cdr().Int() = " << c.Cdr().Int() << endl;
	}
	ExecuteTimer(StressTest, a);
	pause();
	GCCollect();
	ExecuteTimer(IterationTest, a);
	pause();
	ExecuteTimer(StressTest, a);
	pause();
	a = Null;
	GCCollect();
	GCCollect();
	pause();
	return 0;
}

int gcmain2(int argc, _TCHAR* argv[])
{
	UNREFERENCED_PARAMETER(argc);
	UNREFERENCED_PARAMETER(argv);

	for(int i = 0; i < 1000000000; i++) {
		RootPointer a = Cons(Null, Null);
		a.Int() = 0;
		a.Int() ++;
	}
	std::cout << "DING!" << std::endl;
	return 0;
}
