#include "StdAfx.h"
#include "MasterPointer.h"
#include "ConsPair.h"
#include "MemoryRegion.h"
#include "CallocInflatable.h"
#include "LispGC.h"

MemoryRegion::MemoryRegion():
	m_data(NULL),
	m_size(0),
	m_used(0),
	m_prevUsed(0)
{
}

void MemoryRegion::SetData(void * data, size_t size)
{
	m_data = (ConsPair *)data;
	m_size = size / sizeof (ConsPair);
}

ConsPair * MemoryRegion::Cons()
{
	if (m_used < m_size)
	{
		ConsPair * result = &(m_data[m_used]);
		++m_used;
		return result;
	}
	else
	{
		return NULL;
	}
}

void MemoryRegion::Clear()
{
	if (m_prevUsed > m_used)
	{
		size_t currentUsedBytes = RoundUpToPageSize(sizeof(ConsPair) * m_used);
		size_t oldUsedBytes = sizeof(ConsPair) * m_prevUsed;
		if (
			(oldUsedBytes > GetPageSize()) &&
			((oldUsedBytes - GetPageSize()) > currentUsedBytes)
		)
		{
			size_t difference = oldUsedBytes - currentUsedBytes;
			cout << "Decommitting: " << difference << endl;
			UTL_DecommitMemory(m_data + currentUsedBytes,
				difference);
		}
	}
	m_prevUsed = m_used;
	m_used = 0;
}

bool MemoryRegion::Belongs(ConsPair * cons)
{
	return (m_data <= cons) &&
		((m_data + m_size) > cons);
}

size_t MemoryRegion::GetUsed()
{
	return m_used;
}

ConsPair * MemoryRegion::Get(size_t index)
{
	return index < m_used ? &(m_data[index]) : NULL;
}
