#include "StdAfx.h"
#include "MasterPointer.h"
#include "ConsPair.h"
#include "RootPointer.h"
#include "MutableList.h"
#include "LispGC.h"

RootPointer::RootPointer():
	m_ptr(NULL),
	m_next(NULL),
	m_prev(NULL)
{
	GlobalLink();
}

RootPointer::RootPointer(const RootPointer & other):
	m_ptr(other.m_ptr)
{
	GlobalLink();
}

RootPointer::~RootPointer()
{
	GlobalUnlink();
}

RootPointer & RootPointer::operator=(const RootPointer & other)
{
	m_ptr = other.m_ptr;
	return *this;
}

bool RootPointer::IsNull() const
{
	GCAdvance();
	return m_ptr == NULL;
}

RootPointer RootPointer::Car() const
{
	GCAdvance();
	RootPointer result;
	ConsPair * consPair = MasterToCons(m_ptr);
	if (consPair == NULL)
	{
		return result;
	}
	result.m_ptr = consPair->m_car;
	return result;
}

RootPointer RootPointer::Cdr() const
{
	GCAdvance();
	RootPointer result;
	ConsPair * consPair = MasterToCons(m_ptr);
	if (consPair == NULL)
	{
		return result;
	}
	result.m_ptr = consPair->m_cdr;
	return result;
}

int & RootPointer::Int()
{
	GCAdvance();
	return m_ptr->m_ptr->m_intData;
}

double & RootPointer::Double()
{
	GCAdvance();
	return m_ptr->m_ptr->m_doubleData;
}

RootPointer & RootPointer::SetCar(const RootPointer & car)
{
	GCAdvance();
	ConsPair * consPair = m_ptr->m_ptr;
	consPair->m_car = car.IsNull() ? NULL :
		car.m_ptr;
	ScanPointer(this);
	return *this;
}

RootPointer & RootPointer::SetCdr(const RootPointer & cdr)
{
	GCAdvance();
	ConsPair * consPair = m_ptr->m_ptr;
	consPair->m_cdr = cdr.IsNull() ? NULL :
		cdr.m_ptr;
	ScanPointer(this);
	return *this;
}

void RootPointer::GlobalLink()
{
	RootList.Link(this);
}

void RootPointer::GlobalUnlink()
{
	RootList.Unlink(this);
}
